####################################################
#
# SOLVES -Div(2*nu*e(u)) + grad(P) = f, Div(u) = 0
# e(u) = 0.5*(grad(u) + transpose(grad(u)))
###################################################

from netgen.geom2d import SplineGeometry
from ngsolve import *
import numpy as np
from netgen.geom2d import unit_square
import time

geo = SplineGeometry()
#geo.AddRectangle( (0, 0), (2*np.pi, np.pi), bcs = ("wall"))
pnts =[ (0.0,0.0),(2*np.pi,0.0),(2*np.pi,np.pi), (0.0,np.pi)] # Domain Vertices

pnts = [geo.AppendPoint(*pnt) for pnt in pnts]

curves = [[["line",0,1],"wall",1, 0], \
          [["line",1,2],"wall",1, 0],  \
          [["line",2,3],"wall",1, 0], \
          [["line",3,0],"wall",1, 0]] # Domain Edges

for c,bc,l,r in curves:
	geo.Append(c,bc=bc,leftdomain=l, rightdomain=r)
	geo.SetMaterial(1,"Domain")

def outer(W, n):
	return CoefficientFunction((W[0]*n[0], W[0]*n[1], W[1]*n[0], W[1]*n[1]), dims=(2,2)) 

def e(a,b):
	grad_a = grad(a)
	grad_b = grad(b)
	grad_mat = CoefficientFunction((grad_a[0], grad_b[0], grad_a[1],grad_b[1]), dims=(2,2))
	return 0.5*(grad_mat + grad_mat.trans)

def gradient(a,b):
	grad_a = grad(a)
	grad_b = grad(b)
	return CoefficientFunction((grad_a[0], grad_b[0], grad_a[1],grad_b[1]), dims=(2,2))

def poisson(order_p, mesh_size):
	print("Start...")
	mesh = Mesh(geo.GenerateMesh(maxh=mesh_size, quad_dominated=True)) # Mesh Size
	mesh.Curve(3)
	Draw(mesh)

	C = 4.0

	h = specialcf.mesh_size
	
	nu = CoefficientFunction(1.0 + x) # viscosity


	X = L2(mesh,order=order_p)
	fes = FESpace([X,X,X], dgjumps=True)

	ux, uy, p = fes.TrialFunction()
	vx, vy, q = fes.TestFunction()

	n = specialcf.normal(2)
	
	ux_other = ux.Other()
	uy_other = uy.Other()
	p_other = p.Other()
	vx_other = vx.Other()
	vy_other = vy.Other()
	q_other = q.Other()
	

	alpha = C*order_p*order_p / h #PENALTY PARAMETER

	#OPERATORS
	
	U_vec = CoefficientFunction((ux, uy))
	U_Other = CoefficientFunction((ux_other, uy_other))

	V_vec = CoefficientFunction((vx, vy))
	V_Other = CoefficientFunction((vx_other, vy_other))

	jump_U	= U_vec - U_Other
	jump_V = V_vec - V_Other
	
	mean_U = 0.5*(U_vec + U_Other)
	mean_V = 0.5*(V_vec + V_Other)

	mean_gradeU = 0.5*(e(ux, uy) + e(ux.Other(), uy.Other()))
	mean_gradeV = 0.5*(e(vx, vy) + e(vx.Other(), vy.Other()))

	jump_gradU = gradient(ux,uy) - gradient(ux.Other(), uy.Other())
	jump_gradV = gradient(vx, vy) - gradient(vx.Other(), vy.Other())

	jump_q = q - q_other
	jump_p = p - p_other

	jump_2nuinverse_p = (p/(2*nu)) - (p_other/(2*nu.Other()))

	mean_q = 0.5*(q + q_other)
	mean_p = 0.5*(p + p_other)

	jump_2nuU =  2*nu*U_vec - 2*nu.Other()*U_Other
	jump_2nuV =  2*nu*V_vec - 2*nu.Other()*V_Other

	mean_2nu_gradU = 2*0.5 *( nu * gradient(ux,uy) + nu.Other() * gradient(ux.Other(),uy.Other()))
	mean_2nu_gradV = 2*0.5 *( nu * gradient(vx,vy) + nu.Other() * gradient(vx.Other(), vy.Other()))
	mean_2nu_gradeV = 2*0.5 *( nu * e(vx,vy) + nu.Other() * e(vx.Other(),vy.Other()))

	a = BilinearForm(fes)

	#a(u,v)
	a += SymbolicBFI( InnerProduct(e(ux,uy),(2*nu*e(vx,vy))) )

	a += SymbolicBFI(alpha*InnerProduct(outer(jump_U,n),outer(jump_2nuV,n)), VOL, skeleton=True)
	a += SymbolicBFI(alpha*InnerProduct(outer(U_vec,n),outer(2*nu*V_vec,n)), BND, skeleton=True, definedon=mesh.Boundaries("wall"))

	a += SymbolicBFI(-(InnerProduct(outer(jump_U,n), mean_2nu_gradeV) + InnerProduct(outer(jump_2nuV,n), mean_gradeU)), VOL, skeleton=True)
	a += SymbolicBFI(-(InnerProduct(outer(U_vec,n), 2*nu*e(vx,vy)) + InnerProduct(outer(2*nu*V_vec,n),e(ux,uy))),BND, skeleton=True, definedon=mesh.Boundaries("wall"))

	
	#b(v,p)
	a += SymbolicBFI(V_vec*grad(p))
	a += SymbolicBFI(-mean_V*jump_p*n, VOL, skeleton=True)

	#-b(u,q)
	a += SymbolicBFI(-(U_vec*grad(q)))
	a += SymbolicBFI(mean_U*jump_q*n, VOL, skeleton=True)

	#s(p,q)
	a += SymbolicBFI(h*jump_2nuinverse_p*jump_q, VOL, skeleton=True )
	a.Assemble()


	l = LinearForm(fes)

	p_grad_func = CoefficientFunction((cos(x)*sin(y),sin(x)*cos(y)))

	grad_u_b = CoefficientFunction( (-nu*sin(x)*sin(y), nu*cos(x)*cos(y)) )
	grad_v_b = CoefficientFunction( (-nu*cos(x)*cos(y), nu*sin(x)*sin(y)) )


	grad_g_D = CoefficientFunction( (-sin(x)*sin(y), -cos(x)*cos(y),cos(x)*cos(y), sin(x)*sin(y)), dims=(2,2) )
	nu2_e_g_D = 2*nu*0.5*(grad_g_D + grad_g_D.trans)
	grad2_u = CoefficientFunction((nu2_e_g_D[0,0].Diff(x) + nu2_e_g_D[1,0].Diff(y), nu2_e_g_D[0,1].Diff(x) + nu2_e_g_D[1,1].Diff(y) ))
	f_func = CoefficientFunction(-grad2_u + p_grad_func)

	l += SymbolicLFI(InnerProduct(f_func,V_vec))

	#DIRICHLET

	u_b = CoefficientFunction(cos(x)*sin(y))
	v_b = CoefficientFunction(-sin(x)*cos(y))

	u_exact = CoefficientFunction((u_b, v_b))

	l += SymbolicLFI(alpha*u_exact*2*nu*V_vec , BND, skeleton=True, definedon=mesh.Boundaries("wall"))

	l += SymbolicLFI(-InnerProduct(outer(u_exact, n), 2*nu*gradient(vx,vy)), BND, skeleton=True, definedon=mesh.Boundaries("wall"))

	l += SymbolicLFI(-q*n*u_exact, BND, skeleton=True, definedon=mesh.Boundaries("wall"))

	l.Assemble()


	#VISUALIZATION
	gfu = GridFunction(fes)
	

	gfu.vec.data = a.mat.Inverse() * l.vec
	Redraw()
	exact = CoefficientFunction((cos(x)*sin(y),-sin(x)*cos(y)))
	vel = sqrt(gfu.components[0]**2 + gfu.components[1]**2)
	exact_vel = sqrt(u_b*u_b + v_b*v_b)

	pres = CoefficientFunction(sin(x)*sin(y))

	p_bar = 1/(2*np.pi*np.pi) * Integrate(gfu.components[2], mesh)
	Draw(gfu.components[0], mesh, "u_vel")
	Draw(gfu.components[1], mesh, "v_vel")
	Draw(gfu.components[2] - p_bar, mesh, "Pressure")

	h_v = sqrt(Integrate( (vel - exact_vel)*(vel - exact_vel), mesh)) #Velocity error
	h_p = sqrt(Integrate(((gfu.components[2] - p_bar) - sin(x)*sin(y))**2, mesh )) #Pressure Error
	return (h_v, h_p)

from math import log10, floor
def round_sig(x, sig=2):
	return round(x, sig-int(floor(log10(abs(x))))-1)

p = 2
mesh_h = 1/4

h1 = poisson(p, mesh_h) #Choose order & meshsize
time.sleep(2)
h2 = poisson(p, mesh_h/2)

r_v = h1[0]/h2[0]
r_p = h1[1]/h2[1]

conv_v = log(r_v)/log(2)
conv_p = log(r_p)/log(2)

print("V-Error\tP-Error\tV-Conv\tP-Conv")
print(round_sig(h2[0],2), round_sig(h2[1],2),"\t", round_sig(conv_v,2),"\t", round_sig(conv_p,2))



